{
  Demo Multithread

  Application ralise dans le but de montrer comment raliser
  du multithread avec Delphi (version XE2 dans ce cas prcis).

  Ici, le multithread servira  "dcouper" une longue procdure (boucle)
  en multiple oprations simultanes.
  Un ListBox contient plusieurs URLs, et nous souhaitons savoir laquelle
  de ces URLs contient une expression prcise.

  L'opration se prsente ainsi : on prend une URL ; on rcupre son contenu
  par GET via Indy que l'on affecte  une StringList ; on vrifie
  si l'expression choisie est prsente dans la StringList ;
  et ensuite nous attribuons un rsultat positif ou ngatif  l'opration.

  Le multithread est encore tout frais pour moi, ce code est probablement
  optimisable et/ou comporte quelques erreurs :)

  Les commentaires sont  lire dans l'ordre d'excution du programme.
  Chaque commentaire concerne le code du dessous.

  Beny (Delphi forum, Developpez.com)
}

unit Main;

interface

uses
  // Pour synchroniser les threads, vous aurrez besoin de "System.SyncObjs".
  // Pour utiliser Indy (GET), vous aurez besoin de "IdHTTP".
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  Vcl.StdCtrls, Vcl.Samples.Spin, System.SyncObjs, IdHTTP;

type
  TFormMain = class(TForm)
    ListBox: TListBox;
    LabelThreads: TLabel;
    SpinEditThreads: TSpinEdit;
    ButtonStart: TButton;
    ButtonStop: TButton;
    LabelTitre: TLabel;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure ButtonStartClick(Sender: TObject);
    procedure ButtonStopClick(Sender: TObject);
  private
  public
  end;

  // La CLASS de notre thread.
  TDemoThread = class(TThread)
    // Dans PRIVATE nous plaons les variables qui seront modifies
    // par les diffrents threads simultanment, et qui pourront communiquer
    // entre le thread et les synchronisations sans interfrences.
  private
    URL: String;
    LigneActive, Resultat: Integer;
  protected
    // La procdure principal.
    procedure Execute; override;
  public
    // Pour la synchronisation de fin de procdure.
    procedure Finish;
    // Pour la synchronisation de dbut de procdure.
    procedure Initialize;
    // Ce qui lancera chaque thread.
    constructor Create(CreateSuspended: Boolean);
  end;

var
  FormMain: TFormMain;
  CS: TCriticalSection; // La CriticalSection (voir FormCreate plus bas).
  SL: TStringList; // La StringList (voir FormCreate plus bas).
  Threads, Ligne: Integer; // Nous y reviendrons plus tard...
  Work: Boolean; // Quand l'opration commence, WORK passe  TRUE.

implementation

{$R *.dfm}

// Le thread est cre ici.
constructor TDemoThread.Create(CreateSuspended: Boolean);
begin
  inherited Create(CreateSuspended);
end;

// Voici le code principal du thread.
// Vous verrez qu'il n'interagit jamais avec l'interface, cette tape sera
// gre par les synchronisations que nous verrons plus bas.
procedure TDemoThread.Execute;
var
  Raw: TStringList; // StringList temporaire pour stocker le code-source.
  HTTP: TIdHTTP; // On dclare ici le composant HTTP de Indy.
begin
  // Si "WORK=TRUE" l'opration continue normalement...
  while Work do
  begin
    // Ici nous allons vrifier s'il reste des URLs  traiter.
    // Nous allons utiliser ici la CriticalSection afin que cette opration
    // ne soit pas effectue simultanment par plusieurs threads ;
    // ce qui fausserait le rsultat.

    // Nous entrons en section critique...
    CS.Enter;

    // Nous incrmentons de "+1" la position actuelle dans la StringList
    // contenant les URLs  traiter.
    // Pour rappel, LIGNE est un Integer.
    Inc(Ligne);

    // Si la position actuelle est infrieur  la dernire ligne
    // de la StringList, alors nous allons pouvoir traiter cette URL ;
    if Ligne < SL.Count then
      LigneActive := Ligne
    else
      // Sinon, on en dduit qu'il n'y a plus d'URLs  traiter.
      Work := False;

    // Nous quittons la section critique.
    CS.Leave;

    // Si la vrification effectue ci-dessus donne un rsultat positif
    // alors nous pouvons procder  l'opration.
    if Work then
    begin
      // On rcupre l'URL en cours et on l'affecte  un simple String.
      // Ce n'est pas essentiel mais cela limite les connexions entre
      // la StringList contenant les URLs et les threads.
      // Important : comme cette variable sera utilise dans le
      // thread principal (ici) et dans les SYNCHRONIZE, il est ncessaire
      // de dclarer "URL:STRING" dans PRIVATE de la CLASS "TDemoThread".
      URL := SL[LigneActive];

      // On intragit avec l'interface afin que l'utilisateur sache
      // ce qu'il se passe et o nous en sommes.
      Synchronize(Initialize);

      // On cre la StringList temporaire.
      Raw := TStringList.Create;

      // On cre le composant HTTP et on le paramtre vite-fait, bien-fait.
      HTTP := TIdHTTP.Create;
      HTTP.HandleRedirects := True;
      HTTP.ReadTimeout := 15000;
      HTTP.ConnectTimeout := 15000;

      // Dans un TRY/EXCEPT on fait un GET pour rcuprer le code source
      // de l'URL, que l'on va affecter  la StringList cre prcdemment.
      try
        Raw.Text := HTTP.Get(URL);
      except
        // En cas d'erreur, pas de panique (EXCEPT peut rester vide).
        Application.ProcessMessages;
      end;

      // On vrifie si RAW.TEXT (le code-source) contient l'expression "utf-8".
      // Ici nous utilisons POS mais nous pouvions utiliser ANSICONTAINSSTR.
      if Pos('utf-8', Raw.Text) <> 0 then
        Resultat := 1 // La rponse est oui !
      else
        Resultat := 0; // La rponse est non !

      // On libre les composants utiliss via FREE ou FREEANDNIL.
      Raw.Free;
      HTTP.Free;

      // On synchronise la rponse avec la procdure FINISH,
      // qui pourra intragir alors avec l'interface.
      Synchronize(Finish);
    end;
  end;

  // Le thread est termin donc on retire (dcrmente) "-1"  la StringList.
  // Pour rappel, THREADS est un Integer.
  Dec(Threads);

  // Plus de threads dans la liste, on en dduit que l'opration est termine !
  if Threads <= 0 then
    ShowMessage('Traitement complet termin !');
end;

// Nous modifions la ligne de l'URL en cours lorsque l'opraton commence.
procedure TDemoThread.Initialize;
begin
  // Intraction avec l'interface possible,
  // car cette procdure est appele via SYNCHRONIZE.
  FormMain.ListBox.Items.Strings[LigneActive] := URL + ' - En cours...';
end;

// L'opration est termine, place  l'affectation des rsultats.
procedure TDemoThread.Finish;
begin
  case Resultat of
    1: // Rsultat positif, donc l'URL contient l'expression !
      begin
        // Intraction avec l'interface car SYNCHRONIZE :)
        FormMain.ListBox.Items.Strings[LigneActive] :=
          URL + ' - Contient l''expression "utf-8"';
      end;
    0: // Rsultat ngatif, donc l'URL ne contient pas l'expression !
      begin
        // Intraction avec l'interface car SYNCHRONIZE :)
        FormMain.ListBox.Items.Strings[LigneActive] :=
          URL + ' - Ne contient pas l''expression "utf-8"';
      end;
  end;
end;

{ ============================================================================ }

procedure TFormMain.ButtonStartClick(Sender: TObject);
var
  i: Integer;
begin
  // On vrifie que le ListBox contient bien des lments,
  // sinon a risque de ne pas fonctionner...
  // Bien entendu, cette vrification pourrait tre plus pousse.
  if FormMain.ListBox.Items.Count > 0 then
  begin
    // Nous passons WORK  TRUE afin d'indiquer au programme
    // que l'opration est lance !
    Work := True;

    // Si le nombre de threads souhaits est suprieur au nombre d'URLs
    //  traiter, alors on baisse le nombre de threads automatiquement.
    if SpinEditThreads.Value > ListBox.Items.Count then
      SpinEditThreads.Value := ListBox.Items.Count;

    // Afin de ne pas utiliser le ListBox (donc l'interface),
    // nous utiliserons une StringList
    SL.Assign(ListBox.Items);

    // LIGNE (Integer) sera la position actuelle dans la StringList (SL).
    // Nous devons l'initialiser  "-1" car le dbut du thread
    // va l'incrmenter de "+1" (INC) ; LIGNE se doit donc de partir de 0.
    Ligne := -1;

    // Nous indiquons le nombre de threads choisi par l'utilisateur.
    Threads := SpinEditThreads.Value;

    // Voici enfin la cration de chaque thread !
    // Un SPINEDIT est utilis afin de dfinir le nombre de threads maximum.
    for i := 1 to SpinEditThreads.Value do
      TDemoThread.Create(False);
  end
  else
    // ListBox vide, donc on engueule l'utilisateur, c'est normal :D
    ShowMessage('Je te laisse deviner l''erreur...');
end;

procedure TFormMain.ButtonStopClick(Sender: TObject);
begin
  // Si "WORK=TRUE" on en dduit que l'opration est dj en cours...
  if Work then
  begin
    // Sinon, on passe WORK  FALSE afin qu'auncun nouveau thread ne soit cr.
    Work := False;

    // Puis un message indiquant  l'utilisateur que l'opration a t stoppe.
    ShowMessage('Stopp ! Attendons la fin des derniers threads actifs...');
  end
  else
    // ...donc voici un beau message pour en informer l'utilisateur.
    // Il va de soit qu'il est plus lgant de dsactiver les controls
    // inutiles lorsque l'opration est en cours.
    // Mais faisons trs simple ici :)
    ShowMessage('Rien  stopper, l''opration n''a pas commence !');
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  // On libre la StringList avant de fermer.
  SL.Free;

  // On libre la CriticalSection avant de fermer.
  CS.Free;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin
  // On cre la StringList qui servira  stocker de manire virtuelle
  // le contenu du ListBox.
  SL := TStringList.Create;

  // On cre la CriticalSection qui servira  indiquer l'action
  // qui devra tre ralise par les threads un par un,
  // et non simultanment (pour viter les conflits).
  CS := TCriticalSection.Create;
end;

end.
